@@ -1,6 +1,7 @@
 #!/bin/sh
 
 fndir=$DESTDIR$fndir
+scriptdir=$DESTDIR$scriptdir
 
 /bin/sh $sdir_top/mkinstalldirs $fndir || exit 1;
 
@@ -16,23 +17,32 @@ for file in $allfuncs; do
     case "$file" in
       */CVS/*) continue;;
     esac
-    if test x$FUNCTIONS_SUBDIRS != x -a x$FUNCTIONS_SUBDIRS != xno; then
+    if test x$FUNCTIONS_SUBDIRS != x && test x$FUNCTIONS_SUBDIRS != xno; then
       case "$file" in
-      Completion/comp*)
-        subdir="`echo $file | sed -e 's%/[^/]*/[^/]*$%%'`"
-        instdir="$fndir/Completion"
-        ;;
-      Completion/*)
+      Completion/*/*)
         subdir="`echo $file | sed -e 's%/[^/]*/[^/]*$%%'`"
         instdir="$fndir/$subdir"
         ;;
+      Completion/*)
+        instdir="$fndir/Completion"
+        ;;
+      Scripts/*)
+        instdir="$scriptdir"
+	;;
       *)
         subdir="`echo $file | sed -e 's%/[^/]*$%%' -e 's%^Functions/%%'`"
         instdir="$fndir/$subdir"
         ;;
       esac
     else
-      instdir="$fndir"
+      case "$file" in
+      Scripts/*)
+        instdir="$scriptdir"
+	;;
+      *)
+        instdir="$fndir"
+        ;;
+      esac
     fi
     test -d $instdir || /bin/sh $sdir_top/mkinstalldirs $instdir || exit 1
     $INSTALL_DATA $sdir_top/$file $instdir || exit 1
