@@ -67,7 +67,7 @@ bin_strftime(char *nam, char **argv, Options ops, UNUSED(int func))
     }
 
     if (scalar) {
-	setsparam(scalar, ztrdup(buffer));
+	setsparam(scalar, metafy(buffer, -1, META_DUP));
     } else {
 	printf("%s\n", buffer);
     }
